#
# $Id: ebb9798ad0b211e031670a12a1ab154678c1c8f3 $
#
# Copyright (c) 1999-2006 Minero Aoki
#
# This program is free software.
# You can distribute/modify this program under the same terms of ruby.
# see the file "COPYING".

module Racc
  class Error < StandardError; end
  class CompileError < Error; end
end
